import re
from unittest import mock

import pytest

from cognite.client.data_classes import TimeSeries, TimeSeriesFilter, TimeSeriesList, TimeSeriesUpdate
from tests.utils import jsgz_load


@pytest.fixture
def mock_ts_response(rsps, cognite_client):
    response_body = {
        "items": [
            {
                "id": 0,
                "externalId": "string",
                "name": "stringname",
                "isString": True,
                "metadata": {"metadata-key": "metadata-value"},
                "unit": "string",
                "assetId": 0,
                "isStep": True,
                "description": "string",
                "securityCategories": [0],
                "createdTime": 0,
                "lastUpdatedTime": 0,
            }
        ]
    }
    url_pattern = re.compile(
        re.escape(cognite_client.time_series._get_base_url_with_base_path())
        + r"/timeseries(?:/byids|/update|/delete|/list|/search|$|\?.+)"
    )
    rsps.assert_all_requests_are_fired = False

    rsps.add(rsps.POST, url_pattern, status=200, json=response_body)
    rsps.add(rsps.GET, url_pattern, status=200, json=response_body)
    yield rsps


class TestTimeSeries:
    def test_retrieve_single(self, cognite_client, mock_ts_response):
        res = cognite_client.time_series.retrieve(id=1)
        assert isinstance(res, TimeSeries)
        assert mock_ts_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_retrieve_multiple(self, cognite_client, mock_ts_response):
        res = cognite_client.time_series.retrieve_multiple(ids=[1])
        assert isinstance(res, TimeSeriesList)
        assert mock_ts_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_list(self, cognite_client, mock_ts_response):
        res = cognite_client.time_series.list()
        assert mock_ts_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_list_with_filters(self, cognite_client, mock_ts_response):
        res = cognite_client.time_series.list(
            is_step=True,
            is_string=False,
            metadata={"a": "b"},
            last_updated_time={"min": 45},
            created_time={"max": 123},
            asset_ids=[1, 2],
            asset_external_ids=["aeid"],
            data_set_ids=[1, 2],
            data_set_external_ids=["x"],
            root_asset_ids=[1231],
            include_metadata=False,
            asset_subtree_ids=[1],
            asset_subtree_external_ids=["a"],
        )
        assert mock_ts_response.calls[0].response.json()["items"] == res.dump(camel_case=True)
        assert {
            "isString": False,
            "isStep": True,
            "metadata": {"a": "b"},
            "assetIds": [1, 2],
            "assetExternalIds": ["aeid"],
            "assetSubtreeIds": [{"id": 1}, {"externalId": "a"}],
            "dataSetIds": [{"id": 1}, {"id": 2}, {"externalId": "x"}],
            "rootAssetIds": [1231],
            "createdTime": {"max": 123},
            "lastUpdatedTime": {"min": 45},
        } == jsgz_load(mock_ts_response.calls[0].request.body)["filter"]

    @pytest.mark.dsl
    def test_list_with_asset_ids(self, cognite_client, mock_ts_response):
        import numpy

        cognite_client.time_series.list(asset_ids=[1])
        cognite_client.time_series.list(asset_ids=[numpy.int64(1)])
        for i in range(len(mock_ts_response.calls)):
            assert [1] == jsgz_load(mock_ts_response.calls[i].request.body)["filter"]["assetIds"]

    @pytest.mark.dsl
    def test_list_with_root_asset_ids(self, cognite_client, mock_ts_response):
        import numpy

        cognite_client.time_series.list(root_asset_ids=[1])
        cognite_client.time_series.list(root_asset_ids=[numpy.int64(1)])
        for i in range(len(mock_ts_response.calls)):
            assert [1] == jsgz_load(mock_ts_response.calls[i].request.body)["filter"]["rootAssetIds"]

    def test_create_single(self, cognite_client, mock_ts_response):
        res = cognite_client.time_series.create(TimeSeries(external_id="1", name="blabla"))
        assert isinstance(res, TimeSeries)
        assert mock_ts_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_create_multiple(self, cognite_client, mock_ts_response):
        res = cognite_client.time_series.create([TimeSeries(external_id="1", name="blabla")])
        assert isinstance(res, TimeSeriesList)
        assert mock_ts_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_iter_single(self, cognite_client, mock_ts_response):
        for asset in cognite_client.time_series:
            assert mock_ts_response.calls[0].response.json()["items"][0] == asset.dump(camel_case=True)

    def test_iter_chunk(self, cognite_client, mock_ts_response):
        for assets in cognite_client.time_series(chunk_size=1):
            assert mock_ts_response.calls[0].response.json()["items"] == assets.dump(camel_case=True)

    def test_delete_single(self, cognite_client, mock_ts_response):
        res = cognite_client.time_series.delete(id=1)
        assert {"items": [{"id": 1}], "ignoreUnknownIds": False} == jsgz_load(mock_ts_response.calls[0].request.body)
        assert res is None

    def test_delete_single_ignore_unknown(self, cognite_client, mock_ts_response):
        res = cognite_client.time_series.delete(id=1, ignore_unknown_ids=True)
        assert {"items": [{"id": 1}], "ignoreUnknownIds": True} == jsgz_load(mock_ts_response.calls[0].request.body)
        assert res is None

    def test_delete_multiple(self, cognite_client, mock_ts_response):
        res = cognite_client.time_series.delete(id=[1])
        assert {"items": [{"id": 1}], "ignoreUnknownIds": False} == jsgz_load(mock_ts_response.calls[0].request.body)
        assert res is None

    def test_update_with_resource_class(self, cognite_client, mock_ts_response):
        res = cognite_client.time_series.update(TimeSeries(id=1))
        assert isinstance(res, TimeSeries)
        assert mock_ts_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_update_with_update_class(self, cognite_client, mock_ts_response):
        res = cognite_client.time_series.update(TimeSeriesUpdate(id=1).description.set("blabla"))
        assert isinstance(res, TimeSeries)
        assert mock_ts_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_update_multiple(self, cognite_client, mock_ts_response):
        res = cognite_client.time_series.update([TimeSeriesUpdate(id=1).description.set("blabla")])
        assert isinstance(res, TimeSeriesList)
        assert mock_ts_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_update_multiple_list(self, cognite_client, mock_ts_response):
        tsl = cognite_client.time_series.retrieve_multiple(ids=[0])
        res = cognite_client.time_series.update(tsl)
        assert isinstance(res, TimeSeriesList)
        assert 1 == len(res)

    def test_search(self, cognite_client, mock_ts_response):
        res = cognite_client.time_series.search(filter=TimeSeriesFilter(is_string=True))
        assert mock_ts_response.calls[0].response.json()["items"] == res.dump(camel_case=True)
        assert {
            "search": {"name": None, "description": None, "query": None},
            "filter": {"isString": True},
            "limit": 100,
        } == jsgz_load(mock_ts_response.calls[0].request.body)

    @pytest.mark.parametrize("filter_field", ["is_string", "isString"])
    def test_search_dict_filter(self, cognite_client, mock_ts_response, filter_field):
        res = cognite_client.time_series.search(filter={filter_field: True})
        assert mock_ts_response.calls[0].response.json()["items"] == res.dump(camel_case=True)
        assert {
            "search": {"name": None, "description": None, "query": None},
            "filter": {"isString": True},
            "limit": 100,
        } == jsgz_load(mock_ts_response.calls[0].request.body)

    def test_search_with_filter(self, cognite_client, mock_ts_response):
        res = cognite_client.time_series.search(
            name="n", description="d", query="q", filter=TimeSeriesFilter(unit="bla")
        )
        assert mock_ts_response.calls[0].response.json()["items"] == res.dump(camel_case=True)
        req_body = jsgz_load(mock_ts_response.calls[0].request.body)
        assert "bla" == req_body["filter"]["unit"]
        assert {"name": "n", "description": "d", "query": "q"} == req_body["search"]

    def test_update_object(self):
        assert isinstance(
            TimeSeriesUpdate(1)
            .asset_id.set(1)
            .asset_id.set(None)
            .description.set("")
            .description.set(None)
            .external_id.set("1")
            .external_id.set(None)
            .metadata.set({})
            .name.set("")
            .name.set(None)
            .security_categories.set([])
            .unit.set("")
            .unit.set(None),
            TimeSeriesUpdate,
        )
        assert isinstance(
            TimeSeriesUpdate(1)
            .metadata.add({})
            .metadata.remove([])
            .security_categories.add([])
            .security_categories.remove([]),
            TimeSeriesUpdate,
        )


@pytest.mark.dsl
class TestPlotTimeSeries:
    @pytest.fixture
    def mock_get_dps(self, cognite_client, rsps):
        rsps.add(
            rsps.POST,
            cognite_client.time_series._get_base_url_with_base_path() + "/timeseries/data/list",
            status=200,
            json={
                "items": [
                    {
                        "id": 0,
                        "externalId": "string1",
                        "isString": False,
                        "isStep": False,
                        "datapoints": [{"timestamp": i * 10000, "average": i} for i in range(5000)],
                    }
                ]
            },
        )

    @mock.patch("matplotlib.pyplot.show")
    @mock.patch("pandas.core.frame.DataFrame.rename")
    def test_plot_time_series_name_labels(
        self, pandas_rename_mock, plt_show_mock, mock_ts_response, mock_get_dps, cognite_client
    ):
        res = cognite_client.time_series.retrieve(id=0)
        df_mock = mock.MagicMock()
        pandas_rename_mock.return_value = df_mock
        res.plot(aggregates=["average"], granularity="1d")

        assert {0: "stringname", "0|average": "stringname|average"} == pandas_rename_mock.call_args[1]["columns"]
        assert 1 == df_mock.plot.call_count
        assert 1 == plt_show_mock.call_count

    @mock.patch("matplotlib.pyplot.show")
    @mock.patch("pandas.core.frame.DataFrame.plot")
    def test_plot_time_series_id_labels(
        self, pandas_plot_mock, plt_show_mock, mock_ts_response, mock_get_dps, cognite_client
    ):
        res = cognite_client.time_series.retrieve(id=0)
        res.plot(id_labels=True, aggregates=["average"], granularity="1s")

        assert 1 == pandas_plot_mock.call_count
        assert 1 == plt_show_mock.call_count

    @mock.patch("matplotlib.pyplot.show")
    @mock.patch("pandas.core.frame.DataFrame.rename")
    def test_plot_time_series_list_name_labels(
        self, pandas_rename_mock, plt_show_mock, mock_ts_response, mock_get_dps, cognite_client
    ):
        res = cognite_client.time_series.retrieve_multiple(ids=[0])
        df_mock = mock.MagicMock()
        pandas_rename_mock.return_value = df_mock
        res.plot(aggregates=["average"], granularity="1h")
        assert {0: "stringname", "0|average": "stringname|average"} == pandas_rename_mock.call_args[1]["columns"]
        assert 1 == df_mock.plot.call_count
        assert 1 == plt_show_mock.call_count

    @mock.patch("matplotlib.pyplot.show")
    @mock.patch("pandas.core.frame.DataFrame.plot")
    def test_plot_time_series_list_id_labels(
        self, pandas_plot_mock, plt_show_mock, mock_ts_response, mock_get_dps, cognite_client
    ):
        res = cognite_client.time_series.retrieve_multiple(ids=[0])
        res.plot(id_labels=True)

        assert 1 == pandas_plot_mock.call_count
        assert 1 == plt_show_mock.call_count


@pytest.fixture
def mock_time_series_empty(rsps, cognite_client):
    url_pattern = re.compile(re.escape(cognite_client.time_series._get_base_url_with_base_path()) + "/.+")
    rsps.assert_all_requests_are_fired = False
    rsps.add(rsps.POST, url_pattern, status=200, json={"items": []})
    rsps.add(rsps.GET, url_pattern, status=200, json={"items": []})
    yield rsps


@pytest.mark.dsl
class TestPandasIntegration:
    def test_time_series_list_to_pandas(self, cognite_client, mock_ts_response):
        import pandas as pd

        df = cognite_client.time_series.list().to_pandas()
        assert isinstance(df, pd.DataFrame)
        assert 1 == df.shape[0]
        assert {"metadata-key": "metadata-value"} == df["metadata"][0]

    def test_time_series_list_to_pandas_empty(self, cognite_client, mock_time_series_empty):
        import pandas as pd

        df = cognite_client.time_series.list().to_pandas()
        assert isinstance(df, pd.DataFrame)
        assert df.empty

    def test_time_series_to_pandas(self, cognite_client, mock_ts_response):
        import pandas as pd

        df = cognite_client.time_series.retrieve(id=1).to_pandas()
        assert isinstance(df, pd.DataFrame)
        assert "metadata" not in df.columns
        assert [0] == df.loc["securityCategories"][0]
        assert "metadata-value" == df.loc["metadata-key"][0]



import re

from setuptools import find_packages, setup

version = re.search('^__version__\s*=\s*"(.*)"', open("cognite/client/_version.py").read(), re.M).group(1)

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="cognite-sdk",
    version=version,
    description="Client library for Cognite Data Fusion (CDF)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://cognite-sdk-python.readthedocs-hosted.com",
    download_url="https://github.com/cognitedata/cognite-sdk-python/archive/{}.tar.gz".format(version),
    author="Erlend Vollset",
    author_email="erlend.vollset@cognite.com",
    install_requires=["requests>=2.21.0,<3.0.0", "pandas", "requests_oauthlib==1.3.0"],
    python_requires=">=3.5",
    packages=["cognite." + p for p in find_packages(where="cognite")],
    include_package_data=True,
)



import random
import string

import pytest

from cognite.client.data_classes import Transformation, TransformationDestination, TransformationUpdate


@pytest.fixture
def new_transformation(cognite_client):
    prefix = "".join(random.choice(string.ascii_letters) for i in range(6))
    transform = Transformation(
        name="any", external_id=f"{prefix}-transformation", destination=TransformationDestination.assets()
    )
    ts = cognite_client.transformations.create(transform)

    yield ts

    cognite_client.transformations.delete(id=ts.id)
    assert cognite_client.transformations.retrieve(ts.id) is None


other_transformation = new_transformation


class TestTransformationsAPI:
    def test_create_asset_transformation(self, cognite_client):
        prefix = "".join(random.choice(string.ascii_letters) for i in range(6))
        transform = Transformation(
            name="any", external_id=f"{prefix}-transformation", destination=TransformationDestination.assets()
        )
        ts = cognite_client.transformations.create(transform)
        cognite_client.transformations.delete(id=ts.id)

    def test_create_raw_transformation(self, cognite_client):
        prefix = "".join(random.choice(string.ascii_letters) for i in range(6))
        transform = Transformation(
            name="any",
            external_id=f"{prefix}-transformation",
            destination=TransformationDestination.raw("myDatabase", "myTable"),
        )
        ts = cognite_client.transformations.create(transform)
        cognite_client.transformations.delete(id=ts.id)
        assert ts.destination == TransformationDestination.raw("myDatabase", "myTable")

    def test_create_asset_hierarchy_transformation(self, cognite_client):
        prefix = "".join(random.choice(string.ascii_letters) for i in range(6))
        transform = Transformation(
            name="any", external_id=f"{prefix}-transformation", destination=TransformationDestination.asset_hierarchy()
        )
        ts = cognite_client.transformations.create(transform)
        cognite_client.transformations.delete(id=ts.id)

    def test_create_string_datapoints_transformation(self, cognite_client):
        prefix = "".join(random.choice(string.ascii_letters) for i in range(6))
        transform = Transformation(
            name="any",
            external_id=f"{prefix}-transformation",
            destination=TransformationDestination.string_datapoints(),
        )
        ts = cognite_client.transformations.create(transform)
        cognite_client.transformations.delete(id=ts.id)

    def test_create(self, new_transformation):
        assert (
            new_transformation.name == "any"
            and new_transformation.destination == TransformationDestination.assets()
            and new_transformation.id is not None
        )

    def test_retrieve(self, cognite_client, new_transformation):
        retrieved_transformation = cognite_client.transformations.retrieve(new_transformation.id)
        assert (
            new_transformation.name == retrieved_transformation.name
            and new_transformation.destination == retrieved_transformation.destination
            and new_transformation.id == retrieved_transformation.id
        )

    def test_retrieve_multiple(self, cognite_client, new_transformation, other_transformation):
        retrieved_transformations = cognite_client.transformations.retrieve_multiple(
            ids=[new_transformation.id, other_transformation.id]
        )
        assert len(retrieved_transformations) == 2
        assert new_transformation.id in [
            transformation.id for transformation in retrieved_transformations
        ] and other_transformation.id in [transformation.id for transformation in retrieved_transformations]

    def test_update_full(self, cognite_client, new_transformation):
        expected_external_id = f"m__{new_transformation.external_id}"
        new_transformation.external_id = expected_external_id
        new_transformation.name = "new name"
        new_transformation.query = "SELECT * from _cdf.assets"
        new_transformation.destination = TransformationDestination.raw("myDatabase", "myTable")
        updated_transformation = cognite_client.transformations.update(new_transformation)
        retrieved_transformation = cognite_client.transformations.retrieve(new_transformation.id)
        assert (
            updated_transformation.external_id == retrieved_transformation.external_id == expected_external_id
            and updated_transformation.name == retrieved_transformation.name == "new name"
            and updated_transformation.query == retrieved_transformation.query == "SELECT * from _cdf.assets"
            and updated_transformation.destination == TransformationDestination.raw("myDatabase", "myTable")
        )

    def test_update_partial(self, cognite_client, new_transformation):
        expected_external_id = f"m__{new_transformation.external_id}"
        update_transformation = (
            TransformationUpdate(id=new_transformation.id)
            .external_id.set(expected_external_id)
            .name.set("new name")
            .query.set("SELECT * from _cdf.assets")
        )
        updated_transformation = cognite_client.transformations.update(update_transformation)
        retrieved_transformation = cognite_client.transformations.retrieve(new_transformation.id)
        assert (
            updated_transformation.external_id == retrieved_transformation.external_id == expected_external_id
            and updated_transformation.name == retrieved_transformation.name == "new name"
            and updated_transformation.query == retrieved_transformation.query == "SELECT * from _cdf.assets"
        )

    def test_list(self, cognite_client, new_transformation):
        retrieved_transformations = cognite_client.transformations.list()
        assert new_transformation.id in [transformation.id for transformation in retrieved_transformations]

    def test_preview(self, cognite_client):
        query_result = cognite_client.transformations.preview(query="select 1 as id, 'asd' as name", limit=100)
        assert (
            query_result.schema is not None
            and query_result.results is not None
            and len(query_result.schema) == 2
            and len(query_result.results) == 1
            and query_result.results[0]["id"] == 1
            and query_result.results[0]["name"] == "asd"
        )

    def test_preview_to_string(self, cognite_client):
        query_result = cognite_client.transformations.preview(query="select 1 as id, 'asd' as name", limit=100)
        dumped = str(query_result)



import gzip
import json
import re

import pytest

from cognite.client.data_classes import Label, LabelFilter, Relationship, RelationshipList, RelationshipUpdate
from tests.utils import jsgz_load


@pytest.fixture
def mock_rel_response(rsps, cognite_client):
    response_body = {
        "items": [
            {
                "externalId": "string",
                "sourceExternalId": "string",
                "sourceType": "asset",
                "targetExternalId": "string",
                "targetType": "asset",
                "startTime": 0,
                "endTime": 0,
                "confidence": 0,
                "dataSetId": 1,
                "createdTime": 0,
                "lastUpdatedTime": 0,
                "labels": [{"externalId": "string"}],
            }
        ]
    }
    url_pattern = re.compile(
        re.escape(cognite_client.relationships._get_base_url_with_base_path())
        + r"/relationships(?:/byids|/update|/delete|/list|/search|$|\?.+)"
    )
    rsps.assert_all_requests_are_fired = False

    rsps.add(rsps.POST, url_pattern, status=200, json=response_body)
    rsps.add(rsps.GET, url_pattern, status=200, json=response_body)
    yield rsps


@pytest.fixture
def mock_rel_empty(rsps, cognite_client):
    response_body = {"items": []}
    url_pattern = re.compile(
        re.escape(cognite_client.relationships._get_base_url_with_base_path())
        + r"/relationships(?:/byids|/update|/delete|/list|/search|$|\?.+)"
    )
    rsps.assert_all_requests_are_fired = False

    rsps.add(rsps.POST, url_pattern, status=200, json=response_body)
    rsps.add(rsps.GET, url_pattern, status=200, json=response_body)
    yield rsps


class TestRelationships:
    def test_retrieve_single(self, cognite_client, mock_rel_response):
        res = cognite_client.relationships.retrieve(external_id="a")
        assert isinstance(res, Relationship)
        assert mock_rel_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)
        assert {"items": [{"externalId": "a"}], "fetchResources": False} == jsgz_load(
            mock_rel_response.calls[0].request.body
        )

    def test_retrieve_multiple(self, cognite_client, mock_rel_response):
        res = cognite_client.relationships.retrieve_multiple(external_ids=["a"])
        assert isinstance(res, RelationshipList)
        assert mock_rel_response.calls[0].response.json()["items"] == res.dump(camel_case=True)
        assert {"items": [{"externalId": "a"}], "fetchResources": False} == jsgz_load(
            mock_rel_response.calls[0].request.body
        )

    def test_list(self, cognite_client, mock_rel_response):
        res = cognite_client.relationships.list()
        assert mock_rel_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_create_single(self, cognite_client, mock_rel_response):
        res = cognite_client.relationships.create(
            Relationship(
                external_id="1",
                confidence=0.5,
                labels=[Label("belongsTo")],
                source_type="asset",
                source_external_id="source_ext_id",
                target_type="asset",
                target_external_id="bbb",
                data_set_id=12345,
            )
        )
        assert isinstance(res, Relationship)
        assert mock_rel_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_create_single_types(self, cognite_client, mock_rel_response):
        types = ["asset", "timeSeries", "file", "event", "sequence"]
        for cls in types:
            res = cognite_client.relationships.create(
                Relationship(
                    external_id="1",
                    confidence=0.5,
                    labels=[Label("belongsTo")],
                    source_type=cls,
                    source_external_id="source_ext_id",
                    target_type="asset",
                    target_external_id="bbb",
                )
            )
            assert isinstance(res, Relationship)
            res = cognite_client.relationships.create(
                Relationship(
                    external_id="1",
                    confidence=0.5,
                    labels=[Label("belongsTo")],
                    source_type="asset",
                    source_external_id="foo",
                    target_type=cls,
                    target_external_id="bar",
                )
            )
            assert isinstance(res, Relationship)
            res = cognite_client.relationships.create(
                Relationship(
                    external_id="1",
                    confidence=0.5,
                    labels=[Label("belongsTo")],
                    source_type=cls,
                    source_external_id="foo",
                    target_type=cls,
                    target_external_id="bar",
                )
            )
            assert isinstance(res, Relationship)

        for call in mock_rel_response.calls:
            x = json.loads(gzip.decompress(call.request.body).decode("utf-8"))["items"]
            it = x[0]
            assert isinstance(it["sourceType"], str)
            assert isinstance(it["targetType"], str)

    def test_create_wrong_type(self, cognite_client, mock_rel_response):
        with pytest.raises(TypeError):
            cognite_client.relationships.create(
                Relationship(
                    external_id="1",
                    confidence=0.5,
                    labels=[Label("belongsTo")],
                    source_type="relationship",
                    source_external_id="foo",
                    target_type="asset",
                    target_external_id="bar",
                )
            )

    def test_create_multiple(self, cognite_client, mock_rel_response):
        rel1 = Relationship(
            external_id="new1",
            confidence=0.5,
            labels=[Label("flowsTo")],
            source_type="asset",
            source_external_id="foo",
            target_type="asset",
            target_external_id="bar",
        )
        rel2 = Relationship(
            external_id="new2",
            confidence=0.1,
            labels=[Label("flowsTo")],
            source_type="asset",
            source_external_id="foo",
            target_type="asset",
            target_external_id="bar",
        )
        res = cognite_client.relationships.create([rel1, rel2])
        assert isinstance(res, RelationshipList)
        assert mock_rel_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_update_with_resource_class(self, cognite_client, mock_rel_response):
        res = cognite_client.relationships.update(Relationship(external_id="test_1"))
        assert isinstance(res, Relationship)
        assert mock_rel_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_update_with_update_class(self, cognite_client, mock_rel_response):
        res = cognite_client.relationships.update(RelationshipUpdate(external_id="test_1").confidence.set(None))
        assert isinstance(res, Relationship)
        assert mock_rel_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_update_multiple(self, cognite_client, mock_rel_response):
        res = cognite_client.relationships.update(
            [
                RelationshipUpdate(external_id="test1").source_external_id.set("blabla"),
                RelationshipUpdate(external_id="test2").source_external_id.set("blabla"),
            ]
        )
        assert isinstance(res, RelationshipList)
        assert mock_rel_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_update_labels_single(self, cognite_client, mock_rel_response):
        cognite_client.relationships.update(
            [RelationshipUpdate(external_id="test1").labels.add("PUMP").labels.remove("VALVE")]
        )
        expected = {"labels": {"add": [{"externalId": "PUMP"}], "remove": [{"externalId": "VALVE"}]}}
        assert expected == jsgz_load(mock_rel_response.calls[0].request.body)["items"][0]["update"]

    def test_update_labels_multiple(self, cognite_client, mock_rel_response):
        cognite_client.relationships.update(
            [
                RelationshipUpdate(external_id="test1")
                .labels.add(["PUMP", "ROTATING_EQUIPMENT"])
                .labels.remove(["VALVE", "VERIFIED"])
            ]
        )
        expected = {
            "labels": {
                "add": [{"externalId": "PUMP"}, {"externalId": "ROTATING_EQUIPMENT"}],
                "remove": [{"externalId": "VALVE"}, {"externalId": "VERIFIED"}],
            }
        }
        assert expected == jsgz_load(mock_rel_response.calls[0].request.body)["items"][0]["update"]

    # resource.update doesn't support full replacement of labels (set operation)
    def test_ignore_labels_resource_class(self, cognite_client, mock_rel_response):
        cognite_client.relationships.update(
            Relationship(external_id="test1", labels=[Label(external_id="Pump")], source_external_id="source1")
        )
        expected = {"sourceExternalId": {"set": "source1"}}
        assert expected == jsgz_load(mock_rel_response.calls[0].request.body)["items"][0]["update"]

    def test_iter_single(self, cognite_client, mock_rel_response):
        for rel in cognite_client.relationships:
            assert mock_rel_response.calls[0].response.json()["items"][0] == rel.dump(camel_case=True)

    def test_delete_single(self, cognite_client, mock_rel_response):
        res = cognite_client.relationships.delete(external_id="a")
        assert {"items": [{"externalId": "a"}], "ignoreUnknownIds": False} == jsgz_load(
            mock_rel_response.calls[0].request.body
        )
        assert res is None

    def test_delete_multiple(self, cognite_client, mock_rel_response):
        res = cognite_client.relationships.delete(external_id=["a"])
        assert {"items": [{"externalId": "a"}], "ignoreUnknownIds": False} == jsgz_load(
            mock_rel_response.calls[0].request.body
        )
        assert res is None

    def test_delete_multiple_ignore_unknown_ids(self, cognite_client, mock_rel_response):
        res = cognite_client.relationships.delete(external_id=[1], ignore_unknown_ids=True)
        assert {"items": [{"externalId": 1}], "ignoreUnknownIds": True} == jsgz_load(
            mock_rel_response.calls[0].request.body
        )
        assert res is None

    def test_advanced_list(self, cognite_client, mock_rel_response):
        res = cognite_client.relationships.list(
            source_types=["asset"], labels=LabelFilter(contains_any=["label_ext_id"])
        )
        assert {
            "filter": {"sourceTypes": ["asset"], "labels": {"containsAny": [{"externalId": "label_ext_id"}]}},
            "limit": 100,
            "cursor": None,
            "fetchResources": False,
        } == jsgz_load(mock_rel_response.calls[0].request.body)
        assert mock_rel_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_source_target_packing(self, cognite_client, mock_rel_response):
        res = cognite_client.relationships.list(
            source_types=["asset"],
            source_external_ids=["bla"],
            target_types=["timeseries"],
            target_external_ids=["foo"],
            labels=LabelFilter(contains_any=["belongs_to"]),
        )
        assert {
            "filter": {
                "sourceTypes": ["asset"],
                "sourceExternalIds": ["bla"],
                "targetTypes": ["timeseries"],
                "targetExternalIds": ["foo"],
                "labels": {"containsAny": [{"externalId": "belongs_to"}]},
            },
            "limit": 100,
            "cursor": None,
            "fetchResources": False,
        } == jsgz_load(mock_rel_response.calls[0].request.body)
        assert mock_rel_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_multi_source_target_list(self, cognite_client, mock_rel_response):
        source_external_ids = ["source1", "source2"]
        source_types = ["asset", "asset"]
        target_external_ids = ["target1", "target2"]
        target_types = ["event", "event"]
        data_sets = [{"id": 1234}, {"externalId": "test_dataSet_id"}]
        data_set_ids = [1234]
        data_set_external_ids = ["test_dataSet_id"]
        created_time = 1565965333132
        last_updated_time = 1565965333132
        res = cognite_client.relationships.list(
            source_external_ids=source_external_ids,
            source_types=source_types,
            target_external_ids=target_external_ids,
            target_types=target_types,
            data_set_ids=data_set_ids,
            data_set_external_ids=data_set_external_ids,
            created_time=created_time,
            last_updated_time=last_updated_time,
        )
        assert {
            "filter": {
                "createdTime": created_time,
                "lastUpdatedTime": last_updated_time,
                "sourceTypes": source_types,
                "sourceExternalIds": source_external_ids,
                "targetTypes": target_types,
                "targetExternalIds": target_external_ids,
                "dataSetIds": data_sets,
            },
            "limit": 100,
            "cursor": None,
            "fetchResources": False,
        } == jsgz_load(mock_rel_response.calls[0].request.body)
        assert mock_rel_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_many_source_targets(self, cognite_client, mock_rel_response):
        source_external_ids = [str(i) for i in range(2500)]
        target_external_ids = [str(i) for i in range(3500)]
        with pytest.raises(ValueError):
            cognite_client.relationships(
                source_external_ids=source_external_ids, target_external_ids=target_external_ids
            )
        with pytest.raises(ValueError):
            cognite_client.relationships.list(
                source_external_ids=source_external_ids, target_external_ids=target_external_ids
            )
        res = cognite_client.relationships.list(
            source_external_ids=source_external_ids, target_external_ids=target_external_ids, limit=None
        )
        assert 12 == len(mock_rel_response.calls)
        assert isinstance(res, RelationshipList)
        assert 12 == len(res)

    def test_many_sources_only(self, cognite_client, mock_rel_response):
        source_external_ids = [str(i) for i in range(2500)]
        with pytest.raises(ValueError):
            cognite_client.relationships(source_external_ids=source_external_ids)

        res = cognite_client.relationships.list(source_external_ids=source_external_ids, limit=-1)
        assert 3 == len(mock_rel_response.calls)
        assert isinstance(res, RelationshipList)
        assert 3 == len(res)
        requested_sources = []
        for call in mock_rel_response.calls:
            json = jsgz_load(call.request.body)
            assert "targetExternalIds" not in json["filter"]
            requested_sources.extend([s for s in json["filter"]["sourceExternalIds"]])
        assert set([s for s in source_external_ids]) == set(requested_sources)



import doctest
from unittest import TextTestRunner

import pytest

from cognite.client._api import (
    assets,
    data_sets,
    datapoints,
    entity_matching,
    events,
    files,
    iam,
    login,
    raw,
    relationships,
    sequences,
    three_d,
    time_series,
)

# this fixes the issue with 'got MagicMock but expected Nothing in docstrings'
doctest.OutputChecker.__check_output = doctest.OutputChecker.check_output
doctest.OutputChecker.check_output = lambda self, want, got, optionflags: not want or self.__check_output(
    want, got, optionflags
)


def run_docstring_tests(module):
    runner = TextTestRunner()
    s = runner.run(doctest.DocTestSuite(module))
    assert 0 == len(s.failures)


@pytest.mark.usefixtures("mock_cognite_client")
class TestDocstringExamples:
    def test_time_series(self):
        run_docstring_tests(time_series)

    def test_assets(self):
        run_docstring_tests(assets)

    @pytest.mark.dsl
    def test_datapoints(self):
        run_docstring_tests(datapoints)

    def test_data_sets(self):
        run_docstring_tests(data_sets)

    def test_events(self):
        run_docstring_tests(events)

    def test_files(self):
        run_docstring_tests(files)

    def test_login(self):
        run_docstring_tests(login)

    @pytest.mark.dsl
    def test_raw(self):
        run_docstring_tests(raw)

    def test_3d(self):
        run_docstring_tests(three_d)

    def test_iam(self):
        run_docstring_tests(iam)

    def test_sequences(self):
        run_docstring_tests(sequences)

    def test_relationships(self):
        run_docstring_tests(relationships)


@pytest.mark.usefixtures("mock_cognite_beta_client")
class TestDocstringExamplesBeta:
    def test_entity_matching(self):
        run_docstring_tests(entity_matching)



import socket
from unittest.mock import MagicMock

import pytest
import requests.exceptions
import urllib3.exceptions

from cognite.client._http_client import HTTPClient, HTTPClientConfig, _RetryTracker
from cognite.client.exceptions import CogniteConnectionError, CogniteConnectionRefused, CogniteReadTimeout

DEFAULT_CONFIG = HTTPClientConfig(
    status_codes_to_retry={429},
    backoff_factor=0.5,
    max_backoff_seconds=30,
    max_retries_total=10,
    max_retries_read=5,
    max_retries_connect=5,
    max_retries_status=10,
)


class TestRetryTracker:
    def test_total_retries_exceeded(self):
        rt = _RetryTracker(config=DEFAULT_CONFIG)
        rt.config.max_retries_total = 10
        rt.status = 4
        rt.connect = 4
        rt.read = 4

        assert rt.total == 12
        assert rt.should_retry(200) is False

    def test_status_retries_exceeded(self):
        rt = _RetryTracker(config=DEFAULT_CONFIG)
        rt.config.max_retries_status = 1
        assert rt.should_retry(None) is True
        rt.status = 1
        assert rt.should_retry(None) is False

    def test_read_retries_exceeded(self):
        rt = _RetryTracker(config=DEFAULT_CONFIG)
        rt.config.max_retries_read = 1
        assert rt.should_retry(None) is True
        rt.read = 1
        assert rt.should_retry(None) is False

    def test_connect_retries_exceeded(self):
        rt = _RetryTracker(config=DEFAULT_CONFIG)
        rt.config.max_retries_connect = 1
        assert rt.should_retry(None) is True
        rt.connect = 1
        assert rt.should_retry(None) is False

    def test_correct_retry_of_status(self):
        rt = _RetryTracker(config=DEFAULT_CONFIG)
        assert rt.should_retry(500) is False
        rt.config.status_codes_to_retry = {500, 429}
        assert rt.should_retry(500) is True

    def test_get_backoff_time(self):
        rt = _RetryTracker(config=DEFAULT_CONFIG)
        for i in range(1000):
            rt.read += 1
            assert 0 <= rt.get_backoff_time() <= DEFAULT_CONFIG.max_backoff_seconds


def raise_exception_wrapped_as_in_requests_lib(exc: Exception):
    try:
        raise exc
    except type(exc):
        try:
            raise urllib3.exceptions.RequestError(pool=None, url=None, message=None)
        except urllib3.exceptions.RequestError:
            raise requests.exceptions.RequestException


class TestHTTPClient:
    def test_read_timeout_errors(self):
        cnf = DEFAULT_CONFIG
        cnf.max_backoff_seconds = 0
        retry_tracker = _RetryTracker(cnf)
        c = HTTPClient(
            config=cnf,
            retry_tracker_factory=lambda _: retry_tracker,
            session=MagicMock(
                request=MagicMock(
                    side_effect=lambda *args, **kwargs: raise_exception_wrapped_as_in_requests_lib(socket.timeout())
                )
            ),
        )

        with pytest.raises(CogniteReadTimeout):
            c.request("GET", "bla")

        assert retry_tracker.total == DEFAULT_CONFIG.max_retries_read
        assert retry_tracker.read == DEFAULT_CONFIG.max_retries_read
        assert retry_tracker.connect == 0
        assert retry_tracker.status == 0

    @pytest.mark.parametrize("exc_type", [ConnectionAbortedError, ConnectionResetError, BrokenPipeError])
    def test_connect_errors(self, exc_type):
        cnf = DEFAULT_CONFIG
        cnf.max_backoff_seconds = 0
        retry_tracker = _RetryTracker(cnf)
        c = HTTPClient(
            config=cnf,
            retry_tracker_factory=lambda _: retry_tracker,
            session=MagicMock(
                request=MagicMock(
                    side_effect=lambda *args, **kwargs: raise_exception_wrapped_as_in_requests_lib(exc_type())
                )
            ),
        )

        with pytest.raises(CogniteConnectionError):
            c.request("GET", "bla")

        assert retry_tracker.total == DEFAULT_CONFIG.max_retries_connect
        assert retry_tracker.read == 0
        assert retry_tracker.connect == DEFAULT_CONFIG.max_retries_connect
        assert retry_tracker.status == 0

    def test_connection_refused_not_retried(self):
        cnf = DEFAULT_CONFIG
        cnf.max_backoff_seconds = 0
        retry_tracker = _RetryTracker(cnf)
        c = HTTPClient(
            config=cnf,
            retry_tracker_factory=lambda _: retry_tracker,
            session=MagicMock(
                request=MagicMock(
                    side_effect=lambda *args, **kwargs: raise_exception_wrapped_as_in_requests_lib(
                        ConnectionRefusedError()
                    )
                )
            ),
        )

        with pytest.raises(CogniteConnectionRefused):
            c.request("GET", "bla")

        assert retry_tracker.total == 1
        assert retry_tracker.read == 0
        assert retry_tracker.connect == 1
        assert retry_tracker.status == 0

    def test_status_errors(self):
        cnf = DEFAULT_CONFIG
        cnf.max_backoff_seconds = 0
        retry_tracker = _RetryTracker(cnf)
        c = HTTPClient(
            config=cnf,
            retry_tracker_factory=lambda _: retry_tracker,
            session=MagicMock(request=MagicMock(return_value=MagicMock(status_code=429))),
        )

        res = c.request("GET", "bla")
        assert res.status_code == 429

        assert retry_tracker.total == DEFAULT_CONFIG.max_retries_status
        assert retry_tracker.read == 0
        assert retry_tracker.connect == 0
        assert retry_tracker.status == DEFAULT_CONFIG.max_retries_status


class UnderlyingException(Exception):
    pass


class TestGetUnderlyingException:
    def test_get_underlying_exception_does_not_exist_in_context(self):
        try:
            raise Exception
        except Exception as e:
            assert not HTTPClient._any_exception_in_context_isinstance(e, UnderlyingException)

    def test_get_underlying_exception_no_context(self):
        try:
            raise UnderlyingException
        except Exception as e:
            assert HTTPClient._any_exception_in_context_isinstance(e, (UnderlyingException, Exception))

    def test_get_underlying_exception_nested_1_layer(self):
        def testcase():
            try:
                raise UnderlyingException
            except Exception:
                raise Exception()

        try:
            testcase()
        except Exception as e:
            assert HTTPClient._any_exception_in_context_isinstance(e, UnderlyingException)

    def test_get_underlying_exception_nested_2_layers(self):
        def testcase():
            try:
                raise UnderlyingException
            except Exception:
                try:
                    raise Exception()
                except Exception:
                    raise Exception()

        try:
            testcase()
        except Exception as e:
            assert HTTPClient._any_exception_in_context_isinstance(e, UnderlyingException)



import datetime
import logging
import os
import random
import sys
import threading
import types
from multiprocessing.pool import ThreadPool
from time import sleep
from unittest.mock import patch

import pytest

from cognite.client import CogniteClient, utils
from cognite.client._api.assets import AssetList
from cognite.client._api.files import FileMetadataList
from cognite.client._api.time_series import TimeSeriesList
from cognite.client.data_classes import Asset, Event, FileMetadata, TimeSeries
from cognite.client.exceptions import CogniteAPIKeyError
from cognite.client.utils._logging import DebugLogFormatter
from tests.utils import BASE_URL, set_env_var, unset_env_var

_PYPI_ADDRESS = "https://pypi.python.org/simple/cognite-sdk/#history"
TOKEN_URL = "https://test.com/token"


@pytest.fixture
def default_client_config():
    default_config = utils._client_config._DefaultConfig()

    yield (
        "https://greenfield.cognitedata.com",
        default_config.max_workers,
        default_config.timeout,
        "python-sdk-integration-tests",
        "test",
        default_config.token_client_id,
        default_config.token_client_secret,
        default_config.token_url,
        default_config.token_scopes,
    )


@pytest.fixture
def environment_client_config():
    base_url = "blabla"
    num_of_workers = 1
    timeout = 10
    client_name = "test-client"
    project = "test-project"
    client_id = "test-client-id"
    client_secret = "test-client-secret"
    token_url = TOKEN_URL
    token_scopes = "https://test.com/.default,https://test.com/.admin"

    tmp_base_url = os.environ["COGNITE_BASE_URL"]
    tmp_client_name = os.environ["COGNITE_CLIENT_NAME"]
    tmp_project_name = os.getenv("COGNITE_PROJECT")

    os.environ["COGNITE_BASE_URL"] = base_url
    os.environ["COGNITE_MAX_WORKERS"] = str(num_of_workers)
    os.environ["COGNITE_TIMEOUT"] = str(timeout)
    os.environ["COGNITE_CLIENT_NAME"] = client_name
    os.environ["COGNITE_PROJECT"] = project
    os.environ["COGNITE_CLIENT_ID"] = client_id
    os.environ["COGNITE_CLIENT_SECRET"] = client_secret
    os.environ["COGNITE_TOKEN_URL"] = token_url
    os.environ["COGNITE_TOKEN_SCOPES"] = token_scopes

    token_scope_list = token_scopes.split(",")
    yield base_url, num_of_workers, timeout, client_name, project, client_id, client_secret, token_url, token_scope_list

    os.environ["COGNITE_BASE_URL"] = tmp_base_url
    del os.environ["COGNITE_MAX_WORKERS"]
    del os.environ["COGNITE_TIMEOUT"]
    del os.environ["COGNITE_CLIENT_ID"]
    del os.environ["COGNITE_CLIENT_SECRET"]
    del os.environ["COGNITE_TOKEN_URL"]
    del os.environ["COGNITE_TOKEN_SCOPES"]
    os.environ["COGNITE_CLIENT_NAME"] = tmp_client_name

    if tmp_project_name is not None:
        os.environ["COGNITE_PROJECT"] = tmp_project_name
    else:
        del os.environ["COGNITE_PROJECT"]


class TestCogniteClient:
    def test_project_is_correct(self, rsps_with_login_mock):
        with unset_env_var("COGNITE_PROJECT"), set_env_var("COGNITE_API_KEY", "bla"):
            c = CogniteClient()
        assert c.config.project == "test"

    def test_no_api_key_no_token_set(self):
        with unset_env_var("COGNITE_TOKEN_URL", "COGNITE_API_KEY"):
            with pytest.raises(
                CogniteAPIKeyError, match="No API key or token or token generation arguments have been specified"
            ):
                CogniteClient()

    def test_token_factory_set_no_api_key(self):
        c = CogniteClient(token=lambda: "abc")
        assert c.config.token() == "abc"

    def test_token_set_no_api_key(self):
        c = CogniteClient(token="abc")
        assert c.config.token == "abc"

    def test_token_gen_no_api_key(self, rsps, environment_client_config):
        rsps.add(rsps.GET, "https://pypi.python.org/simple/cognite-sdk/#history", status=200, body="")
        rsps.add(
            rsps.POST,
            TOKEN_URL,
            status=200,
            json={"access_token": "abc", "expires_at": datetime.datetime.now().timestamp() + 1000},
        )

        c = CogniteClient()

        assert c.config.token() == "abc"

    def test_token_factory_set_no_api_key_and_no_project(self, rsps_with_login_mock):
        with unset_env_var("COGNITE_PROJECT"):
            c = CogniteClient(token=lambda: "abc")
        assert c.config.project == "test"

    def test_invalid_api_key(self, rsps):
        rsps.add(rsps.GET, _PYPI_ADDRESS, status=200, body="")
        rsps.add(
            rsps.GET,
            BASE_URL + "/login/status",
            status=200,
            json={"data": {"project": "", "loggedIn": False, "user": "", "projectId": -1}},
        )
        with unset_env_var("COGNITE_PROJECT"), set_env_var("COGNITE_API_KEY", "invalid"):
            with pytest.raises(CogniteAPIKeyError):
                CogniteClient()

    def test_no_client_name(self):
        with unset_env_var("COGNITE_CLIENT_NAME"), set_env_var("COGNITE_API_KEY", "bla"):
            with pytest.raises(ValueError, match="No client name has been specified"):
                CogniteClient()

    def assert_config_is_correct(
        self,
        config,
        base_url,
        max_workers,
        timeout,
        client_name,
        project,
        token_client_id,
        token_client_secret,
        token_url,
        token_scopes,
    ):
        assert config.base_url == base_url
        assert type(config.base_url) is str

        assert config.max_workers == max_workers
        assert type(config.max_workers) is int

        assert config.timeout == timeout
        assert type(config.timeout) is int

        assert config.client_name == client_name
        assert type(config.client_name) is str

        assert config.project == project
        assert type(config.project) is str

        assert config.token_client_id == token_client_id
        assert config.token_client_secret == token_client_secret
        assert config.token_url == token_url
        assert config.token_scopes == token_scopes

    def test_default_config(self, rsps_with_login_mock, default_client_config):
        with unset_env_var("COGNITE_PROJECT"), set_env_var("COGNITE_API_KEY", "bla"):
            client = CogniteClient()
        self.assert_config_is_correct(client.config, *default_client_config)

    def test_parameter_config(self):
        base_url = "blabla"
        max_workers = 1
        timeout = 10
        client_name = "test-client"
        project = "something"
        token_client_id = "test-client-id"
        token_client_secret = "test-client-secret"
        token_url = "https://param-test.com/token"
        token_scopes = ["test-scope", "second-test-scope"]
        with set_env_var("COGNITE_API_KEY", "bla"):
            client = CogniteClient(
                project=project,
                base_url=base_url,
                max_workers=max_workers,
                timeout=timeout,
                client_name=client_name,
                token_client_id=token_client_id,
                token_client_secret=token_client_secret,
                token_url=token_url,
                token_scopes=token_scopes,
            )
        self.assert_config_is_correct(
            client._config,
            base_url,
            max_workers,
            timeout,
            client_name,
            project,
            token_client_id,
            token_client_secret,
            token_url,
            token_scopes,
        )

    def test_environment_config(self, environment_client_config):
        with set_env_var("COGNITE_API_KEY", "bla"):
            client = CogniteClient()
        self.assert_config_is_correct(client._config, *environment_client_config)

    @pytest.fixture
    def thread_local_credentials_module(self):
        credentials_module = types.ModuleType("cognite._thread_local")
        credentials_module.credentials = threading.local()
        sys.modules["cognite._thread_local"] = credentials_module
        yield
        del sys.modules["cognite._thread_local"]

    def create_client_and_check_config(self, i):
        from cognite._thread_local import credentials

        api_key = "thread-local-api-key{}".format(i)
        project = "thread-local-project{}".format(i)

        credentials.api_key = api_key
        credentials.project = project

        sleep(random.random())
        client = CogniteClient()

        assert api_key == client.config.api_key
        assert project == client.config.project

    def test_create_client_thread_local_config(self, thread_local_credentials_module):
        with ThreadPool() as pool:
            pool.map(self.create_client_and_check_config, list(range(16)))

    def test_client_debug_mode(self, rsps_with_login_mock):
        with unset_env_var("COGNITE_PROJECT"), set_env_var("COGNITE_API_KEY", "bla"):
            CogniteClient(debug=True)
        log = logging.getLogger("cognite-sdk")
        assert isinstance(log.handlers[0].formatter, DebugLogFormatter)
        log.handlers = []
        log.propagate = False

    def test_version_check_disabled_env(self, rsps_with_login_mock):
        rsps_with_login_mock.assert_all_requests_are_fired = False
        with unset_env_var("COGNITE_PROJECT"), set_env_var("COGNITE_API_KEY", "bla"):
            with set_env_var("COGNITE_DISABLE_PYPI_VERSION_CHECK", "1"):
                CogniteClient()
        assert len(rsps_with_login_mock.calls) == 1
        assert rsps_with_login_mock.calls[0].request.url.startswith("https://greenfield.cognitedata.com")

    def test_version_check_disabled_arg(self, rsps_with_login_mock):
        rsps_with_login_mock.assert_all_requests_are_fired = False
        with unset_env_var("COGNITE_PROJECT", "COGNITE_DISABLE_PYPI_VERSION_CHECK"), set_env_var(
            "COGNITE_API_KEY", "bla"
        ):
            CogniteClient(disable_pypi_version_check=True)
        assert len(rsps_with_login_mock.calls) == 1
        assert rsps_with_login_mock.calls[0].request.url.startswith("https://greenfield.cognitedata.com")

    def test_api_version_present_in_header(self, rsps_with_login_mock, default_client_config):
        with set_env_var("COGNITE_API_KEY", "bla"):
            c = CogniteClient()
        c.login.status()
        assert rsps_with_login_mock.calls[1].request.headers["cdf-version"] == c.config.api_subversion

    def test_beta_header_for_beta_client(self, rsps_with_login_mock, default_client_config):
        from cognite.client.beta import CogniteClient as BetaClient

        with set_env_var("COGNITE_API_KEY", "bla"):
            c = BetaClient()
        c.login.status()
        assert rsps_with_login_mock.calls[1].request.headers["cdf-version"] == "beta"

    def test_version_check_enabled(self, rsps_with_login_mock):
        with unset_env_var("COGNITE_PROJECT"), set_env_var("COGNITE_API_KEY", "bla"):
            CogniteClient()
        assert len(rsps_with_login_mock.calls) == 2

    @patch("cognite.client.utils._version_checker.re.findall")
    @patch("cognite.client.utils._version_checker.requests")
    def test_verify_ssl_enabled_by_default(self, mock_requests, mock_findall):
        with set_env_var("COGNITE_API_KEY", "bla"):
            c = CogniteClient()

        mock_requests.get.assert_called_with(_PYPI_ADDRESS, verify=True, timeout=5)
        assert c._api_client._http_client_with_retry.session.verify is True
        assert c._api_client._http_client.session.verify is True

    @patch("cognite.client.utils._version_checker.re.findall")
    @patch("cognite.client.utils._version_checker.requests")
    def test_verify_ssl_disabled(self, mock_requests, mock_findall):
        with set_env_var("COGNITE_DISABLE_SSL", "1"):
            with set_env_var("COGNITE_API_KEY", "bla"):
                CogniteClient()
            mock_requests.get.assert_called_with(_PYPI_ADDRESS, verify=False, timeout=5)


class TestInstantiateWithClient:
    @pytest.mark.parametrize("cls", [Asset, Event, FileMetadata, TimeSeries])
    def test_instantiate_resources_with_cognite_client(self, cls):
        with set_env_var("COGNITE_API_KEY", "bla"):
            c = CogniteClient()
        assert cls(cognite_client=c)._cognite_client == c

    @pytest.mark.parametrize("cls", [AssetList, Event, FileMetadataList, TimeSeriesList])
    def test_intantiate_resource_lists_with_cognite_client(self, cls):
        with set_env_var("COGNITE_API_KEY", "bla"):
            c = CogniteClient()
        assert cls([], cognite_client=c)._cognite_client == c



import re

import pytest

from cognite.client.data_classes import DataSet, DataSetList, DataSetUpdate, TimestampRange
from tests.utils import jsgz_load


@pytest.fixture
def dataset_update():
    yield DataSetUpdate(id=1).description.set("_")


@pytest.fixture
def mock_ds_response(rsps, cognite_client):
    response_body = {
        "items": [
            {
                "id": 0,
                "externalId": "string",
                "name": "string",
                "description": "string",
                "metadata": {"metadata-key": "metadata-value"},
                "writeProtected": False,
                "createdTime": 0,
                "lastUpdatedTime": 0,
            }
        ]
    }
    url_pattern = re.compile(
        re.escape(cognite_client.data_sets._get_base_url_with_base_path())
        + r"/datasets(?:/byids|/update|/delete|/list|$|\?.+)"
    )
    rsps.assert_all_requests_are_fired = False

    rsps.add(rsps.POST, url_pattern, status=200, json=response_body)
    rsps.add(rsps.GET, url_pattern, status=200, json=response_body)
    yield rsps


class TestDataset:
    def test_retrieve_single(self, cognite_client, mock_ds_response):
        res = cognite_client.data_sets.retrieve(id=1)
        assert isinstance(res, DataSet)
        assert mock_ds_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_retrieve_multiple(self, cognite_client, mock_ds_response):
        res = cognite_client.data_sets.retrieve_multiple(ids=[1])
        assert isinstance(res, DataSetList)
        assert mock_ds_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_list_with_timestamp_range(self, cognite_client, mock_ds_response):
        cognite_client.data_sets.list(created_time=TimestampRange(min=20))
        assert 20 == jsgz_load(mock_ds_response.calls[0].request.body)["filter"]["createdTime"]["min"]
        assert "max" not in jsgz_load(mock_ds_response.calls[0].request.body)["filter"]["createdTime"]

    def test_list_with_time_dict(self, cognite_client, mock_ds_response):
        cognite_client.data_sets.list(last_updated_time={"max": 20})
        assert 20 == jsgz_load(mock_ds_response.calls[0].request.body)["filter"]["lastUpdatedTime"]["max"]
        assert "min" not in jsgz_load(mock_ds_response.calls[0].request.body)["filter"]["lastUpdatedTime"]

    def test_call_root(self, cognite_client, mock_ds_response):
        list(cognite_client.data_sets.__call__(write_protected=True, limit=10))
        calls = mock_ds_response.calls
        assert 1 == len(calls)
        assert {"cursor": None, "limit": 10, "filter": {"writeProtected": True}} == jsgz_load(calls[0].request.body)

    def test_create_single(self, cognite_client, mock_ds_response):
        res = cognite_client.data_sets.create(DataSet(external_id="1", name="blabla"))
        assert isinstance(res, DataSet)
        assert mock_ds_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_create_multiple(self, cognite_client, mock_ds_response):
        res = cognite_client.data_sets.create([DataSet(external_id="1")])
        assert isinstance(res, DataSetList)
        assert mock_ds_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_iter_single(self, cognite_client, mock_ds_response):
        for dataset in cognite_client.data_sets:
            assert mock_ds_response.calls[0].response.json()["items"][0] == dataset.dump(camel_case=True)

    @pytest.mark.skip("delete not implemented")
    def test_delete_single(self, cognite_client, mock_ds_response):
        res = cognite_client.data_sets.delete(id=1)
        assert {"items": [{"id": 1}]} == jsgz_load(mock_ds_response.calls[0].request.body)
        assert res is None

    @pytest.mark.skip("delete not implemented")
    def test_delete_multiple(self, cognite_client, mock_ds_response):
        res = cognite_client.data_sets.delete(id=[1])
        assert {"items": [{"id": 1}]} == jsgz_load(mock_ds_response.calls[0].request.body)
        assert res is None

    def test_update_with_resource_class(self, cognite_client, mock_ds_response):
        res = cognite_client.data_sets.update(DataSet(id=1))
        assert isinstance(res, DataSet)
        assert mock_ds_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_update_with_update_class(self, cognite_client, mock_ds_response, dataset_update):
        res = cognite_client.data_sets.update(dataset_update)
        assert isinstance(res, DataSet)
        assert mock_ds_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_update_multiple(self, cognite_client, mock_ds_response, dataset_update):
        res = cognite_client.data_sets.update([dataset_update])
        assert isinstance(res, DataSetList)
        assert mock_ds_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_event_update_object(self):

        assert isinstance(
            DataSetUpdate(1)
            .description.set("")
            .description.set(None)
            .external_id.set("")
            .external_id.set(None)
            .metadata.set({"_": "."})
            .metadata.set(None)
            .write_protected.set(False)
            .write_protected.set(None)
            .name.set("")
            .name.set(None),
            DataSetUpdate,
        )


@pytest.fixture
def mock_ds_empty(rsps, cognite_client):
    url_pattern = re.compile(re.escape(cognite_client.data_sets._get_base_url_with_base_path()) + "/.+")
    rsps.add(rsps.POST, url_pattern, status=200, json={"items": []})
    yield rsps


@pytest.mark.dsl
class TestPandasIntegration:
    def test_datasets_list_to_pandas(self, cognite_client, mock_ds_response):
        import pandas as pd

        df = cognite_client.data_sets.list().to_pandas()
        assert isinstance(df, pd.DataFrame)
        assert 1 == df.shape[0]
        assert {"metadata-key": "metadata-value"} == df["metadata"][0]

    def test_datasets_list_to_pandas_empty(self, cognite_client, mock_ds_empty):
        import pandas as pd

        df = cognite_client.data_sets.list().to_pandas()
        assert isinstance(df, pd.DataFrame)
        assert df.empty

    def test_datasets_to_pandas(self, cognite_client, mock_ds_response):
        import pandas as pd

        df = cognite_client.data_sets.retrieve(id=1).to_pandas()
        assert isinstance(df, pd.DataFrame)
        assert "metadata" not in df.columns
        assert False == df.loc["writeProtected"].bool()
        assert "metadata-value" == df.loc["metadata-key"][0]
